/**
 * SECTION:element-gstcrMosaic
 * ORG SECTION:element-gstcrseimux
 */


#include <gst/gst.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include <gst/video/video-format.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include "gstnvivameta_api.h"
#include "gst-nvquery.h"
#include "gstcrmosaic.h"
GST_DEBUG_CATEGORY_STATIC (crmosaic_debug);
#define GST_CAT_DEFAULT crmosaic_debug

#define SINK_CAPS "ANY"
#define SRC_CAPS "ANY"

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
		GST_PAD_SRC,
		GST_PAD_ALWAYS,
		GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE
				("{NV12}"))
);
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink_%u",
		GST_PAD_SINK,
		GST_PAD_REQUEST,
		GST_STATIC_CAPS (SINK_CAPS)
);

enum
{
	PROP_0,
	PROP_MOSAIC_TOP,
	PROP_MOSAIC_LEFT,
	PROP_MOSAIC_WIDTH,
	PROP_MOSAIC_HEIGHT,
	PROP_MOSAIC_ACTIVE,
	////////
	PROP_RO_STATUS,
};

#define DEFAULT_OUTPUT_WIDTH (1920)
#define DEFAULT_OUTPUT_HEIGHT (1080)
#define DEFAULT_MOSAIC_TOP	(0)
#define DEFAULT_MOSAIC_LEFT	(1440)
#define DEFAULT_MOSAIC_WIDTH	(480)
#define DEFAULT_MOSAIC_HEIGHT	(270)

/**********************/
G_DEFINE_TYPE (CRMOSAIC, gst_crmosaic, GST_TYPE_ELEMENT);

static GstPad *gst_crmosaic_request_new_pad (GstElement * element,
		GstPadTemplate * templ, const gchar * name, const GstCaps * caps);
static void gst_crmosaic_release_pad (GstElement * element, GstPad * pad);
static GstFlowReturn gst_crmosaic_chain (GstPad * pad, GstObject * parent,
		GstBuffer * buffer);
static gboolean gst_crmosaic_sink_event (GstPad * pad, GstObject * parent,
		GstEvent * event);
static GstStateChangeReturn gst_crmosaic_change_state (GstElement *
		element, GstStateChange transition);
static void gst_crmosaic_set_property (GObject * object, guint prop_id,
		const GValue * value, GParamSpec * pspec);
static void gst_crmosaic_get_property (GObject * object, guint prop_id,
		GValue * value, GParamSpec * pspec);
static void gst_crmosaic_dispose (GObject * object);
static gboolean
gst_crmosaic_src_event (GstPad * pad, GstObject * parent, GstEvent * event);
static gboolean 
gst_crmosaic_src_event_real (CRMOSAIC *crmosaic, GstEvent * event);

static gboolean
gst_crmosaic_setcaps (GstPad * pad, CRMOSAIC * crmosaic, GstCaps * caps)
{
	GstStructure *structure;
	gboolean ret = FALSE;
	GstCaps *peercaps;

	if (caps == NULL)
		return FALSE;

	if (!gst_caps_is_fixed (caps))
		return FALSE;

	peercaps = gst_pad_peer_query_caps (crmosaic->srcpad, NULL);
	if (peercaps) {
		GstCaps *tcaps, *othercaps;;
		tcaps = gst_pad_get_pad_template_caps (pad);
		othercaps = gst_caps_intersect_full (peercaps, tcaps,
				GST_CAPS_INTERSECT_FIRST);

		if (gst_caps_get_size (othercaps) > 0) {
			structure = gst_caps_get_structure (othercaps, 0);
		}

		gst_caps_unref (othercaps);
		gst_caps_unref (peercaps);
		gst_caps_unref (tcaps);
	}

	structure = gst_caps_get_structure (caps, 0);

	if (!structure)
		return FALSE;

	if (crmosaic->send_stream_start) {
		gchar s_id[32];
		/* stream-start (FIXME: create id based on input ids) */
		g_snprintf (s_id, sizeof (s_id), "interleave-%08x", g_random_int ());
		gst_pad_push_event (crmosaic->srcpad, gst_event_new_stream_start (s_id));

		crmosaic->send_stream_start = FALSE;
	}

	GST_DEBUG_OBJECT (crmosaic,
			"setting caps %" GST_PTR_FORMAT " on src pad..", caps);
	ret = gst_pad_set_caps (crmosaic->srcpad, caps);
	gst_caps_unref (caps);

	return ret;
}

/////////////////////////////////////////////
static void
gst_crmosaic_init (CRMOSAIC * crmosaic)
{
	GstElementClass *klass = GST_ELEMENT_GET_CLASS (crmosaic);
	crmosaic->counter = 0;
	crmosaic->mosaic_top = DEFAULT_MOSAIC_TOP;
	crmosaic->mosaic_left = DEFAULT_MOSAIC_LEFT;
	crmosaic->mosaic_width = DEFAULT_MOSAIC_WIDTH;
	crmosaic->mosaic_height = DEFAULT_MOSAIC_HEIGHT;
	crmosaic->active = FALSE;
	crmosaic->batch_size = 1;
	cudaSetDevice(crmosaic->gpu_id);
	NvBufSurfaceCreateParams create_params;
	create_params.gpuId = crmosaic->gpu_id;
	create_params.width = DEFAULT_OUTPUT_WIDTH;
	create_params.height = DEFAULT_OUTPUT_HEIGHT;
	create_params.size = 0;
	create_params.colorFormat = NVBUF_COLOR_FORMAT_NV12;
	create_params.layout = NVBUF_LAYOUT_PITCH;
#ifdef __aarch64__
	create_params.memType = NVBUF_MEM_DEFAULT;
#else
	create_params.memType = NVBUF_MEM_CUDA_UNIFIED;
#endif
	for (int i= 0 ;i < 2;i++) {
		if (NvBufSurfaceCreate(&crmosaic->surface[i], 1, &create_params) != 0) {
			GST_ERROR("Error: Could not allocate internal buffer for params");
		}
	}

	crmosaic->srcpad =
			gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
					"src"), "src");
	gst_pad_set_event_function (crmosaic->srcpad,
			GST_DEBUG_FUNCPTR (gst_crmosaic_src_event));
	gst_pad_use_fixed_caps (crmosaic->srcpad);
	gst_element_add_pad (GST_ELEMENT (crmosaic), crmosaic->srcpad);

	crmosaic->last_stop = GST_CLOCK_TIME_NONE;
}

/////////////////////////////////////////////
static void
gst_crmosaic_class_init (CRMOSAICClass * klass)
{
	GObjectClass *gobject_class;
	GstElementClass *gstelement_class;

	gobject_class = (GObjectClass *) klass;
	gstelement_class = (GstElementClass *) klass;

	///////////////////////////////
	gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_crmosaic_get_property);
	gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_crmosaic_set_property);
	gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_crmosaic_dispose);
	klass->src_event = GST_DEBUG_FUNCPTR (gst_crmosaic_src_event_real);

	//////////////////////////
	gstelement_class->change_state = GST_DEBUG_FUNCPTR (gst_crmosaic_change_state);
	gstelement_class->request_new_pad =
			GST_DEBUG_FUNCPTR (gst_crmosaic_request_new_pad);
	gstelement_class->release_pad = GST_DEBUG_FUNCPTR (gst_crmosaic_release_pad);

	gst_element_class_add_static_pad_template (gstelement_class, &src_factory);
	gst_element_class_add_static_pad_template (gstelement_class, &sink_factory);

	gst_element_class_set_static_metadata (gstelement_class, "CRMOSAIC mixer",
			"crmosaic",
			"Mix multiple video streams with high performance",
			"wush <wush@cr.com> ");
	/////////////////////////////////////////////////////////////////
	g_object_class_install_property(gobject_class, PROP_MOSAIC_TOP,
			g_param_spec_int("mosaic-top", "Mosaic top",
					"Top position of the mix buffer", 0, G_MAXINT,
					DEFAULT_MOSAIC_TOP,
					(GParamFlags) (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));
	g_object_class_install_property(gobject_class, PROP_MOSAIC_LEFT,
			g_param_spec_int("mosaic-left", "Mosaic left",
					"Left position of the mix buffer", 0, G_MAXINT,
					DEFAULT_MOSAIC_LEFT,
					(GParamFlags) (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));
	g_object_class_install_property(gobject_class, PROP_MOSAIC_WIDTH,
			g_param_spec_int("mosaic-width", "Mosaic width",
					"Width of the mix buffer", 1, G_MAXINT,
					DEFAULT_MOSAIC_WIDTH,
					(GParamFlags) (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));
	g_object_class_install_property(gobject_class, PROP_MOSAIC_HEIGHT,
			g_param_spec_int("mosaic-height", "Mosaic height",
					"Height of the mix buffer", 1, G_MAXINT,
					DEFAULT_MOSAIC_HEIGHT,
					(GParamFlags) (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));
	g_object_class_install_property(gobject_class, PROP_MOSAIC_ACTIVE,
			g_param_spec_boolean("active", "active",
					"Activate mixing multiple video streams ",
					FALSE,
					(GParamFlags) (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)));
}

static void
gst_crmosaic_get_property (GObject * object,
		guint prop_id, GValue * value, GParamSpec * pspec)
{
	CRMOSAIC *params = GST_CRMOSAIC (object);

	switch (prop_id) {
	case PROP_MOSAIC_TOP:
		g_value_set_int(value, params->mosaic_top);
		break;
	case PROP_MOSAIC_LEFT:
		g_value_set_int(value, params->mosaic_left);
		break;
	case PROP_MOSAIC_WIDTH:
		g_value_set_int(value, params->mosaic_width);
		break;
	case PROP_MOSAIC_HEIGHT:
		g_value_set_int(value, params->mosaic_height);
		break;
	case PROP_MOSAIC_ACTIVE:
		g_value_set_boolean(value, params->active);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
gst_crmosaic_set_property (GObject * object,
		guint prop_id, const GValue * value, GParamSpec * pspec)
{
	CRMOSAIC *params = GST_CRMOSAIC (object);

	switch (prop_id) {
	case PROP_MOSAIC_TOP:
		params->mosaic_top = g_value_get_int(value);
		break;
	case PROP_MOSAIC_LEFT:
		params->mosaic_left = g_value_get_int(value);
		break;
	case PROP_MOSAIC_WIDTH:
		params->mosaic_width = g_value_get_int(value);
		break;
	case PROP_MOSAIC_HEIGHT:
		params->mosaic_height = g_value_get_int(value);
		break;
	case PROP_MOSAIC_ACTIVE:
		params->active = g_value_get_boolean(value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
gst_crmosaic_dispose (GObject * object)
{
	CRMOSAIC *crmosaic = GST_CRMOSAIC (object);
	GList *item;
	int sinkpadcnt=0;

	GST_DEBUG_OBJECT (crmosaic, " enter dispose");
	g_clear_object (&crmosaic->last_pad);

	for (item = GST_ELEMENT_PADS (object); item; item = g_list_next (item)) {
		GstPad *pad = GST_PAD (item->data);
		if (GST_PAD_IS_SINK (pad)) {
			sinkpadcnt++;
			gst_element_release_request_pad (GST_ELEMENT (object), pad);
			GST_DEBUG_OBJECT (crmosaic, " release sink pad %d", sinkpadcnt);
		}
	}
	for (int i = 0;i < 2;i ++)
		if (crmosaic->surface[i])
			NvBufSurfaceDestroy(crmosaic->surface[i]);

	G_OBJECT_CLASS (gst_crmosaic_parent_class)->dispose (object);
}

static void
gst_crmosaic_ready_to_paused (CRMOSAIC * crmosaic)
{
	GST_OBJECT_LOCK (crmosaic);

	g_clear_object (&crmosaic->last_pad);
	crmosaic->send_stream_start = TRUE;

	GST_OBJECT_UNLOCK (crmosaic);
}

static GstStateChangeReturn
gst_crmosaic_change_state (GstElement * element, GstStateChange transition)
{
	CRMOSAIC *crmosaic = GST_CRMOSAIC (element);
	GstStateChangeReturn ret;

	switch (transition) {
	case GST_STATE_CHANGE_READY_TO_PAUSED:
		gst_crmosaic_ready_to_paused (crmosaic);
		break;
	default:
		break;
	}
	ret = GST_ELEMENT_CLASS (gst_crmosaic_parent_class)->change_state (element,
			transition);
	switch (transition) {
	case GST_STATE_CHANGE_PAUSED_TO_READY:
		g_clear_object (&crmosaic->last_pad);
		break;
	default:
		break;
	}
	return ret;
}

static gboolean
gst_crmosaic_src_event (GstPad * pad, GstObject * parent, GstEvent * event)
{
	CRMOSAIC *crmosaic = GST_CRMOSAIC (parent);
	CRMOSAICClass *klass;
	gboolean ret;

	klass = GST_CRMOSAIC_GET_CLASS (parent);

	ret = klass->src_event (crmosaic, event);

	return ret;
}

static gboolean
gst_crmosaic_src_event_real (CRMOSAIC *crmosaic, GstEvent * event)
{
	switch (GST_EVENT_TYPE (event)) {
	case GST_EVENT_CUSTOM_UPSTREAM:
		break;
	default:
		break;
	}
	return gst_pad_event_default (crmosaic->srcpad, GST_OBJECT (crmosaic), event);
}

static void
gst_crmosaic_setup_sinkpad (CRMOSAIC * crmosaic, GstPad * sinkpad)
{
	/* setup some pad functions */
	gst_pad_set_chain_function (sinkpad, GST_DEBUG_FUNCPTR (gst_crmosaic_chain));
	gst_pad_set_event_function (sinkpad,
			GST_DEBUG_FUNCPTR (gst_crmosaic_sink_event));

	gst_pad_set_active (sinkpad, TRUE);
	gst_element_add_pad (GST_ELEMENT (crmosaic), sinkpad);
}

static GstPad *
gst_crmosaic_request_new_pad (GstElement * element,
		GstPadTemplate * templ, const gchar * req_name, const GstCaps * caps)
{
	CRMOSAIC *crmosaic = GST_CRMOSAIC (element);
	GstPad *newpad;

	if (templ->direction != GST_PAD_SINK) {
		GST_WARNING_OBJECT (crmosaic, "request pad that is not a SINK pad");
		return NULL;
	}

	/* create new pad with the name */
	newpad = gst_pad_new_from_template (templ, req_name);
	GST_INFO_OBJECT (crmosaic, "Adding pad %s", GST_PAD_NAME (newpad));
	if (newpad)
		gst_crmosaic_setup_sinkpad (crmosaic, newpad);
	else
		GST_WARNING_OBJECT (crmosaic, "failed to create request pad");

	return newpad;
}

static void
gst_crmosaic_release_pad (GstElement * element, GstPad * pad)
{
	CRMOSAIC *crmosaic = GST_CRMOSAIC (element);
	GST_INFO_OBJECT (crmosaic, "Remove pad %s", GST_PAD_NAME (pad));
	gst_element_remove_pad (element, pad);
}

static GstFlowReturn
gst_crmosaic_chain (GstPad * pad, GstObject * parent, GstBuffer * buffer)
{
	CRMOSAIC *crmosaic = GST_CRMOSAIC (parent);
	GstFlowReturn flow = GST_FLOW_OK;

	crmosaic->tsmp = GST_BUFFER_TIMESTAMP(buffer);
	crmosaic->duration = GST_BUFFER_DURATION(buffer);

	GST_DEBUG_OBJECT (crmosaic, " pad %s indata len=%lu ts=%lu dur=%lu. ",
			GST_PAD_NAME (pad), gst_buffer_get_size(buffer), crmosaic->tsmp, crmosaic->duration);
	if(strcmp (GST_PAD_NAME (pad), "sink_1" ) == 0)
	{
		GstMapInfo in_map_info;
		gst_buffer_map(buffer, &in_map_info,GST_MAP_READ);

		NvBufSurfTransformParams transform_params;
		NvBufSurfTransform_Error err;
		NvBufSurfTransformRect src_rect;
		NvBufSurfTransformRect dst_rect;
		NvBufSurface ip_surf;
		ip_surf = *(NvBufSurface *)in_map_info.data;

		// Set the transform ROIs for sinter_bufource and destination
		src_rect = {0, 0, 1920, 1080};
		dst_rect = {0, 0, 1920, 1080};

		// Set the transform parameters
		transform_params.src_rect = &src_rect;
		transform_params.dst_rect = &dst_rect;
		transform_params.transform_flag = NVBUFSURF_TRANSFORM_FILTER
				| NVBUFSURF_TRANSFORM_CROP_SRC | NVBUFSURF_TRANSFORM_CROP_DST;
		transform_params.transform_filter = NvBufSurfTransformInter_Default;

		// Transformation scaling+format conversion if any.
		err = NvBufSurfTransform(&ip_surf, crmosaic->surface[0], &transform_params);
		if (err != NvBufSurfTransformError_Success) {
			GST_ELEMENT_ERROR(crmosaic, STREAM, FAILED,
					("NvBufSurfTransform failed with error %d while converting buffer", err), (NULL));
		}
		gst_buffer_unmap(buffer, &in_map_info);

		gst_buffer_unref (buffer);
		crmosaic->main_sub++;
		crmosaic->counter ++;
		return flow;
	}

	if (gst_pad_check_reconfigure (crmosaic->srcpad)) {
		GstCaps *current_caps = gst_pad_get_current_caps (pad);
		if (!gst_crmosaic_setcaps (pad, crmosaic, current_caps)) {
			gst_pad_mark_reconfigure (crmosaic->srcpad);
			gst_buffer_unref (buffer);
		}
		gst_caps_unref (current_caps);
	}
	if ( (crmosaic->counter > 10) && crmosaic->active) {
		GstMapInfo out_map_info, in_map_info;
		NvBufSurfTransformRect src_rect, dest_rect;
		NvBufSurface *surface;
		/* Create a 'outbuf' to instead of 'buffer' */
		GstBuffer * outbuf;
		GstMemory *memory;
		GstAllocationParams params = { (GstMemoryFlags)2, 1, 0, 0};
		guint	outsize = 1920*1080*3/2;
		outbuf = gst_buffer_new ();
		//outbuf = gst_buffer_new_allocate (NULL, outsize, &params);
		memory = gst_allocator_alloc (NULL, outsize, &params);
		gst_buffer_insert_memory (outbuf, -1, memory);
		if (outbuf) {
			gst_buffer_memset(outbuf,0,0,outsize);
		GST_BUFFER_TIMESTAMP (outbuf) = crmosaic->tsmp;
		GST_BUFFER_DURATION (outbuf) = crmosaic->duration;
		gst_buffer_map(outbuf, &out_map_info,GST_MAP_READ);
		///////////////////////////////////////
		gst_buffer_map(buffer, &in_map_info,GST_MAP_READ);
		surface = (NvBufSurface *)in_map_info.data;

		// Map the buffer so that it can be accessed by CPU
		if (NvBufSurfaceMap (surface, -1, -1, NVBUF_MAP_WRITE) != 0) {
			g_assert(false);
		}
		src_rect = {0, 0, 1920, 1080};
		dest_rect = {(guint)crmosaic->mosaic_top, (guint)crmosaic->mosaic_left, \
				(guint)crmosaic->mosaic_width, (guint)crmosaic->mosaic_height};
		crmosaic->composite_params.src_comp_rect = &src_rect;
		crmosaic->composite_params.dst_comp_rect = &dest_rect;
		crmosaic->composite_params.composite_flag = NVBUFSURF_TRANSFORM_COMPOSITE;
		crmosaic->composite_params.input_buf_count = 1;
		if (NvBufSurfTransformComposite(crmosaic->surface[0], surface,
				&crmosaic->composite_params)) {
			g_assert(false);
		}
		// Cache the mapped data for CPU access
		NvBufSurfaceSyncForCpu (surface, -1, -1);
		memcpy(out_map_info.data, surface->surfaceList[0].mappedAddr.addr[0], outsize);
		NvBufSurfaceUnMap (surface, -1, -1);

		GST_WARNING_OBJECT (crmosaic, "in_map_info->size [%d] / outsize [%d]", in_map_info.size, surface->surfaceList[0].dataSize);
		gst_buffer_unmap(outbuf, &out_map_info);
		gst_buffer_unmap(buffer, &in_map_info);

			GstFlowReturn ret = gst_pad_push (crmosaic->srcpad, outbuf);
			if(GST_FLOW_OK != ret)
				GST_ERROR_OBJECT(crmosaic, " pad %s push buf--- failed    [%d]", \
						GST_PAD_NAME (crmosaic->srcpad), ret);
		}
		return flow;
	}

	if(GST_FLOW_OK != gst_pad_push (crmosaic->srcpad, buffer))
		GST_ERROR_OBJECT(crmosaic, " pad %s push buf failed", \
				GST_PAD_NAME (crmosaic->srcpad));
	return flow;
}

static gboolean
gst_crmosaic_sink_event (GstPad * pad, GstObject * parent, GstEvent * event)
{
	gboolean ret;
	CRMOSAIC *crmosaic = GST_CRMOSAIC (parent);

	GST_OBJECT_LOCK (crmosaic);
	gboolean is_pad = (pad == crmosaic->last_pad);
	GST_OBJECT_UNLOCK (crmosaic);

	GST_DEBUG_OBJECT (crmosaic, " enter sink_event by pad %s", GST_PAD_NAME (pad));
	switch (GST_EVENT_TYPE (event)) {
	case GST_EVENT_CAPS:
	{
		GstCaps * caps;
		gst_event_parse_caps (event, &caps);
		GST_INFO_OBJECT (crmosaic, " Received caps-event with caps: %"
				GST_PTR_FORMAT, caps);
		ret = gst_pad_event_default (pad, parent, event);
	}
	break;
	case GST_EVENT_FLUSH_STOP:
		GST_INFO_OBJECT (crmosaic, " Received FLUSH_STOP");
		break;
	case GST_EVENT_SEGMENT:
		GST_INFO_OBJECT (crmosaic, " Received SEGMENT");
		if (is_pad) {
			GstSegment new_segment;
			gst_segment_init (&new_segment, GST_FORMAT_TIME);
			event = gst_event_new_segment (&new_segment);
		}
		break;
	default:
		ret = gst_pad_event_default (pad, parent, event);
		break;
	}
	return ret;
}

gboolean
plugin_init (GstPlugin * plugin)
{
	GST_DEBUG_CATEGORY_INIT (crmosaic_debug, "crmosaic", 0, "CRMOSAIC muxer element");

	return gst_element_register (plugin, "crmosaic", GST_RANK_NONE,
			GST_TYPE_CRMOSAIC);
}

GST_PLUGIN_DEFINE (
		GST_VERSION_MAJOR,
		GST_VERSION_MINOR,
		crmosaic,
		"CRMOSAIC muxer plugin",
		plugin_init,
		VERSION,
		GST_LICENSE,
		GST_PACKAGE_NAME,
		GST_PACKAGE_ORIGIN)
