/*
 * Copyright (c) 2015-2017, NVIDIA CORPORATION.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA Corporation is strictly prohibited.
 *
 */

/**
 * @file
 * <b>NVIDIA GStreamer IVA: Metadata Extension</b>
 *
 * @b Description: This file specifies the NVIDIA IVA Metadata structures used to
 * describe metadata objects.
 */

/**
 * @defgroup gstreamer_utilities_group GStreamer Utilities
 * Defines GStreamer utilities.
 */

/**
 * @defgroup gstreamer_metadata_group GStreamer IVA Metadata Extension
 *
 * Defines an API for managing GStreamer IVA metadata.
 * @ingroup gstreamer_utilities_group
 * @{
 *
 * IVA Metadata is attached to a buffer with gst_buffer_add_iva_meta() or gst_buffer_add_iva_meta_full().
 *
 * Multiple metadatas may be attached by different elements.
 * gst_buffer_get_iva_meta() gets the last added @ref IvaMeta.
 * To iterate through each IvaMeta, following snippet can be used.
 *
 * @code
 *  gpointer state = NULL;
 *  GstMeta *gst_meta;
 *  IvaMeta *ivameta;
 *  static GQuark _ivameta_quark = 0;
 *
 *  if (!_ivameta_quark)
 *    _ivameta_quark = g_quark_from_static_string("ivameta");
 *
 *  while ((gst_meta = gst_buffer_iterate_meta (buf, &state))) {
 *    if (gst_meta_api_type_has_tag(gst_meta->info->api, _ivameta_quark)) {
 *      ivameta = (IvaMeta *) gst_meta;
 *      // Do something with this ivameta
 *    }
 *  }
 * @endcode
 *
 *  The meta_data member of the IvaMeta structure must be cast to a meaningful structure pointer
 *  based on the meta_type. For example, for `meta_type = NV_BBOX_INFO`,
 *  meta_data must be cast as `(BBOX_Params *)`.
 */

#ifndef GST_NVIVA_META_API_H
#define GST_NVIVA_META_API_H

#include <gst/gst.h>
#include <gst/video/video.h>
#include <gst/base/gstbasetransform.h>

#include "nv_object_colors.h"
#include "nvosd.h"

#ifdef __cplusplus
extern "C"
{
#endif

GType iva_meta_api_get_type (void);
#define IVA_META_API_TYPE (iva_meta_api_get_type())

const GstMetaInfo *iva_meta_get_info (void);
#define IVA_META_INFO (iva_meta_get_info())

#define NV_IVA_META_MAX_RECTS NVOSD_MAX_NUM_RECTS
#define MAX_CLR_STRING_LENGTH 32
#define MAX_NUM_CLASSES 32
#define MAX_SEC_LABELS 16

/** Defines GStreamer IVA metadata types. */
typedef enum
{
    NV_META_INIT = 0x0,
    NV_BBOX_DECODE_INFO,
    NV_ENCODE_MV_INFO,
    NV_META_STREAM_ID,
    NV_BBOX_INFO = 0x100,
    NV_BBOX_CROP,
} Meta_type;

/** Holds metadata that contains raw crop image. A `meta_type` member of @ref IvaMeta
    must be set to `NV_BBOX_CROP`. */
typedef struct {
  /** Width of the crop, in pixels. */
  gint width;
  /** Height of the crop, in pixels. */
  gint height;
  /** Color format of the crop; valid only when `is_jpeg` is FALSE. */
  GstVideoFormat clr_format;
  /** Size of the data allocated to crop_buffer, in bytes. */
  gint length;

  /** Pointer to the crop data buffer. */
  gchar *crop_buffer;
  /** Unique tracking ID of the object to which this crop is to be associated. */
  gint tracking_id;

  /** Boolean indicating whether crop_buffer contains jpeg data. */
  gboolean is_jpeg;
} NvBBOXCrop;

/** @cond UNUSED */

/**
 * Holds the box parameters of frame.
 */
typedef struct _RectDim {
  int left;
  int top;
  int width;
  int height;
  int class_id;
  int tracking_id;
  int gie_unique_id;
  int bDetectColor;
  void *text;
}RectDim;

typedef struct _RectInfo {
  RectDim rectDim[NV_IVA_META_MAX_RECTS];
  unsigned int filled_rects;
  unsigned int available_rects;
}RectInfo;
/** @endcond */


/** Holds data that a secondary classifier updates with secondary label information,
 *  such as car type, car color, etc.
 *  The structure may contain a string label. It may also contain a set of
 *  N(num_labels) pairs of `<label_id,label_output>`. For single label classifiers,
 *  `label_id` will always be 0, and N=1. For multi-label classifers, `label_id`
 *  will be the index of the label type (e.g. 0 for Car type, 1 for Car make, ...).
 *  In both cases, `label_output` will be the index of the output of the label type
 *  0 for Sedan, 1 for Coupe, ...)
 */
typedef struct _LabelInfo {
  /** Boolean indicating whether @a str_label is valid. */
  int is_str_label;
  /** String label */
  char str_label[64];
  /** An array of label ids, only @a num_labels elements are valid. */
  gshort label_ids[64];
  /** An array of label outputs, only @a num_labels elements are valid. */
  gshort label_outputs[64];
  /** Number of valid elements in the @a label_ids and @a label_outputs array. */
  int num_labels;
}LabelInfo;

/** Holds parameters that describe metadata for one object. */
typedef struct _ROIMeta_Params {
  /** Structure containing the positional parameters of the object in the frame.
   *  Can also be used to overlay borders / semi-transparent boxes on top of objects
   *  Refer NvOSD_RectParams from nvosd.h
   */
  NvOSD_RectParams rect_params;
  /** Text describing the object can be overlayed using this structure.
   *  @see NvOSD_TextParams from nvosd.h. */
  NvOSD_TextParams text_params;
  /** Index of the object class infered by the primary detector/classifier */
  int class_id;
  /** Used internally by components. */
  int parent_class_id;
  /* Unique ID for tracking the object. */
  int tracking_id;
  /** Deprecated. */
  int bDetectColor;
  /** Deprecated. */
  NvObjectColor object_color;
  /** Deprecated. */
  char detected_clr[MAX_CLR_STRING_LENGTH];
  /** Secondary classifiers update this structure with secondary classification labels.
   *  Each secondary classifier will only update the label_info structure at index specified by the
   *  classifiers unique gie id.
   */
  LabelInfo label_info[MAX_SEC_LABELS];
  /** Used internally by components. */
  struct _ROIMeta_Params *parent_roi;
  /** Boolean indicating whether label_info contains new information. */
  gboolean has_new_info;
  /** Boolean indicating whether secondary classifiers should run inference on the object again. Used internally by components. */
  gboolean force_reinference;

  /** Portion of width of the object overlapping with other object of same type, in pixels. */
  unsigned int occlusion_width;
  /** Portion of height of the object overlapping with other object of same type, in pixels. */
  unsigned int occlusion_height;
} ROIMeta_Params;

/** Holds data that describes metadata objects in the current frame.
    `meta_type` member of @ref IvaMeta must be set to `NV_BBOX_INFO`. */
typedef struct _BBOX_Params {
  /** Array of ROIMeta_Params structure describing each object. */
  ROIMeta_Params *roi_meta;
  /** Number of rectangles/objects i.e. length of @ref ROIMeta_Params */
  unsigned int num_rects;
  /** Number of valid strings in @ref ROIMeta_Params. */
  unsigned int num_strings;
  /** Used internally by components. */
  unsigned int num_labels;
  /** Used internally by components. */
  unsigned int available_rects;
  /** Used internally by components. */
  unsigned int available_texts;
  /** Used internally by components. */
  unsigned int available_labels;
  /** Used internally by components. */
  unsigned int num_classes;
  /** NvOSD mode to be used. @see NvOSD_Mode in `nvosd.h`. */
  NvOSD_Mode nvosd_mode;
  /** Will always be equal to 1 for primary detector. */
  int gie_type;
  /** Batch size of the primary detector. */
  int gie_batch_size;
  /** Used internally by components. Product of batch_size and interval of primary detector.*/
  int gie_batch_x_interval;
  /** Unique ID of the primary detector that attached this metadata. */
  int gie_unique_id;
  /** Frame number. */
  int frame_num;
}BBOX_Params;

typedef struct _IvaMeta IvaMeta;

/** Holds IVA metadata. */
struct _IvaMeta {
  GstMeta       meta;
  /** Must be cast to another structure based on @a meta_type. */
  void *meta_data;
  /** Type of metadata, from the @ref Meta_type enum. */
  int meta_type;
  /** Function called with meta_data pointer as argument when the meta is going to be destroyed.
   * Can be used to clear/free meta_data.
   * Refer to https://developer.gnome.org/glib/unstable/glib-Datasets.html#GDestroyNotify */
  GDestroyNotify destroy;
};

/**
 * Adds GstMeta of type @ref IvaMeta to the GstBuffer and sets the `meta_data` member of @ref IvaMeta.
 *
 * @param[in] buffer GstBuffer to add metadata to.
 * @param[in] meta_data The pointer to set the `meta_data` member of @ref IvaMeta.
 *
 * @return A pointer to the attached @ref IvaMeta structure, NULL in case of failure
 */
IvaMeta *gst_buffer_add_iva_meta (GstBuffer *buffer, void *meta_data);

/**
 * Adds GstMeta like gst_buffer_add_iva_meta(), but this method also has a destroy callback to free
 * attached metadata.
 *
 * @param[in] buffer GstBuffer to which the function adds metadata.
 * @param[in] meta_data The pointer to which the function sets the meta_data member of @ref IvaMeta.
 * @param[in] destroy The GDestroyNotify function to be called when IvaMeta is to be destroyed.
 *            The function is called with meta_data as a parameter. Refer to
 *            https://developer.gnome.org/glib/unstable/glib-Datasets.html#GDestroyNotify.
 *
 * @return A pointer to the attached @ref IvaMeta structure; or NULL in case of failure.
 */
IvaMeta *gst_buffer_add_iva_meta_full (GstBuffer *buffer, void *meta_data,
        GDestroyNotify destroy);

/** Gets the @ref IvaMeta last added to the GstBuffer.
 *
 * @param[in] buffer GstBuffer
 *
 * @return A pointer to the last added IvaMeta structure; or NULL if no IvaMeta was attached.
 */
IvaMeta *gst_buffer_get_iva_meta (GstBuffer *buffer);

/** @} */

typedef enum SingleTrackState
{
	START_NEW,
	KEEP_NOW,
	STOP_TRACK,
	TRACK_STATE
}_TrackState;


#ifdef __cplusplus
}
#endif

#endif
